import os

from io import open


def make_case_function(left_filename):
    right_filename = left_filename.replace('.left.', '.right.')
    expected_filename = left_filename.replace('.left.', '.expected.')

    def test(self):
        with open(expected_filename, 'rt', encoding='utf8') as input_file:
            expected_xml = input_file.read()

        try:
            result_xml = self.process(left_filename, right_filename)
        except Exception as err:
            if u'.err' not in left_filename:
                raise
            result_xml = u'%s: %s' % (err.__class__.__name__, err)

        self.assertEqual(expected_xml.strip(), result_xml.strip())

    return test


def generate_filebased_cases(data_dir, test_class, suffix='xml', ignore=()):
    for left_filename in os.listdir(data_dir):
        if not left_filename.endswith('.left.' + suffix):
            continue
        if left_filename in ignore:
            continue

        left_filename = os.path.join(data_dir, left_filename)
        test_function = make_case_function(left_filename)
        function_name = os.path.split(left_filename)[-1].replace('.', '-')
        test_name = 'test_' + function_name
        setattr(test_class, test_name, test_function)


def compare_elements(left, right):
    path = left.getroottree().getpath(left)
    assert left.text == right.text, "Texts differ: %s" % path
    assert left.tail == right.tail, "Tails differ: %s" % path
    assert left.attrib == right.attrib, "Attributes differ: %s" % path
    # We intentionally do NOT compare namespaces, they are allowed to differ
    assert len(left) == len(right), "Children differ: %s" % path
    for l, r in zip(left.getchildren(), right.getchildren()):
        compare_elements(l, r)



import os
import unittest

from io import open
from lxml import etree
from xmldiff import utils
from xmldiff.diff import (Differ, UpdateTextIn, InsertNode, MoveNode,
                          DeleteNode, UpdateAttrib, InsertAttrib, RenameAttrib,
                          DeleteAttrib, UpdateTextAfter, RenameNode,
                          InsertComment)

from .testing import compare_elements


class APITests(unittest.TestCase):
    left = u"<document><p>Text</p><p>More</p></document>"
    right = u"<document><p>Tokst</p><p>More</p></document>"
    lefttree = etree.fromstring(left)
    righttree = etree.fromstring(right)
    differ = Differ()

    def test_set_trees(self):
        # Passing in just one parameter causes an error:
        with self.assertRaises(TypeError):
            self.differ.set_trees(self.lefttree, None)

        # Passing in something that isn't iterable also cause errors...
        with self.assertRaises(TypeError):
            self.differ.set_trees(object(), self.righttree)

        # This is the way:
        self.differ.set_trees(self.lefttree, self.righttree)

    def test_match(self):
        # Passing in just one parameter causes an error:
        with self.assertRaises(TypeError):
            self.differ.match(self.lefttree, None)

        # Passing in something that isn't iterable also cause errors...
        with self.assertRaises(TypeError):
            self.differ.match(object(), self.righttree)

        # This is the way:
        res1 = self.differ.match(self.lefttree, self.righttree)
        lpath = self.differ.left.getroottree().getpath
        rpath = self.differ.right.getroottree().getpath
        res1x = [(lpath(x[0]), rpath(x[1]), x[2]) for x in res1]

        # Or, you can use set_trees:
        self.differ.set_trees(self.lefttree, self.righttree)
        res2 = self.differ.match()
        lpath = self.differ.left.getroottree().getpath
        rpath = self.differ.right.getroottree().getpath
        res2x = [(lpath(x[0]), rpath(x[1]), x[2]) for x in res2]

        # The match sequences should be the same, of course:
        self.assertEqual(res1x, res2x)
        # But importantly, they are not the same object, meaning the
        # matching was redone.
        self.assertIsNot(res1, res2)
        # However, if we call match() a second time without setting
        # new sequences, we'll get a cached result:
        self.assertIs(self.differ.match(), res2)

    def test_diff(self):
        # Passing in just one parameter causes an error:
        with self.assertRaises(TypeError):
            list(self.differ.diff(self.lefttree, None))

        # Passing in something that isn't iterable also cause errors...
        with self.assertRaises(TypeError):
            list(self.differ.diff(object(), self.righttree))

        # This is the way:
        res1 = list(self.differ.diff(self.lefttree, self.righttree))

        # Or, you can use set_trees() or match()
        # We need to reparse self.lefttree, since after the diffing they
        # are equal.
        self.lefttree = etree.fromstring(self.left)
        self.differ.set_trees(self.lefttree, self.righttree)
        res2 = list(self.differ.diff())

        # The match sequences should be the same, of course:
        self.assertEqual(res1, res2)
        # But importantly, they are not the same object, meaning the
        # matching was redone.
        self.assertIsNot(res1, res2)
        # There is no caching of diff(), so running it again means another
        # diffing.
        self.assertIsNot(list(self.differ.diff()), res2)


class NodeRatioTests(unittest.TestCase):

    def test_compare_equal(self):
        xml = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        tree = etree.fromstring(xml)
        differ = Differ()
        differ.set_trees(tree, tree)
        differ.match()

        # Every node in these trees should get a 1.0 leaf_ratio,
        # and if it has children, 1.0 child_ration, else None
        for left, right in zip(utils.post_order_traverse(differ.left),
                               utils.post_order_traverse(differ.right)):
            self.assertEqual(differ.leaf_ratio(left, right), 1.0)
            if left.getchildren():
                self.assertEqual(differ.child_ratio(left, right), 1.0)
            else:
                self.assertIsNone(differ.child_ratio(left, right))

    def test_compare_different_leafs(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="2" single-ref="2">
            <para>This doesn't match at all</para>
        </section>
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>Completely different from before</para>
        </section>
        <section xml:id="oldfirst" ref="4" single-ref="4">
            <para>Another paragraph</para>
        </section>
        <section ref="5" single-ref="5">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        lefttree = etree.fromstring(left)
        righttree = etree.fromstring(right)
        differ = Differ()

        # Make some choice comparisons here
        # These node are exactly the same
        left = lefttree.xpath('/document/story/section[3]/para')[0]
        right = righttree.xpath('/document/story/section[3]/para')[0]

        self.assertEqual(differ.leaf_ratio(left, right), 1.0)

        # These nodes have slightly different text, but no children
        left = lefttree.xpath('/document/story/section[2]/para')[0]
        right = righttree.xpath('/document/story/section[2]/para')[0]

        self.assertAlmostEqual(differ.leaf_ratio(left, right),
                               0.75)

        # These nodes should not be very similar
        left = lefttree.xpath('/document/story/section[1]/para')[0]
        right = righttree.xpath('/document/story/section[1]/para')[0]
        self.assertAlmostEqual(differ.leaf_ratio(left, right),
                               0.45614035087719)

    def test_compare_different_nodes(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="2" single-ref="2">
            <para>First paragraph</para>
            <para>Second paragraph</para>
        </section>
        <section ref="3" single-ref="3">
            <para>Third paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="2" single-ref="2">
            <para>First paragraph</para>
        </section>
        <section single-ref="3" ref="3">
            <para>Second paragraph</para>
            <para>Third paragraph</para>
        </section>
        <section single-ref="4" ref="4" >
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        differ = Differ()
        differ.set_trees(etree.fromstring(left), etree.fromstring(right))
        differ.match()

        # Make some choice comparisons here. leaf_ratio will always be 1.0,
        # as these leafs have the same attributes and no text, even though
        # attributes may be in different order.
        left = differ.left.xpath('/document/story/section[1]')[0]
        right = differ.right.xpath('/document/story/section[1]')[0]

        self.assertEqual(differ.leaf_ratio(left, right), 1.0)
        # Only one of two matches:
        self.assertEqual(differ.child_ratio(left, right), 0.5)

        left = differ.left.xpath('/document/story/section[2]')[0]
        right = differ.right.xpath('/document/story/section[2]')[0]

        self.assertEqual(differ.leaf_ratio(left, right), 1.0)
        # Only one of two matches:
        self.assertEqual(differ.child_ratio(left, right), 0.5)

        # These nodes should not be very similar
        left = differ.left.xpath('/document/story/section[3]')[0]
        right = differ.right.xpath('/document/story/section[3]')[0]
        self.assertEqual(differ.leaf_ratio(left, right), 1.0)
        self.assertEqual(differ.child_ratio(left, right), 1.0)

    def test_compare_with_xmlid(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="1" single-ref="1">
            <para>First paragraph</para>
            <para>This is the second paragraph</para>
        </section>
        <section ref="3" single-ref="3" xml:id="tobedeleted">
            <para>Det tredje stycket</para>
        </section>
        <section xml:id="last" ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="newfirst" ref="1" single-ref="1">
            <para>First paragraph</para>
        </section>
        <section xml:id="oldfirst" single-ref="2" ref="2">
            <para>This is the second</para>
            <para>Det tredje stycket</para>
        </section>
        <section single-ref="4" ref="4" >
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        differ = Differ()
        differ.set_trees(etree.fromstring(left), etree.fromstring(right))
        differ.match()

        # Make some choice comparisons here.

        left = differ.left.xpath('/document/story/section[1]')[0]
        right = differ.right.xpath('/document/story/section[1]')[0]

        # These are very similar
        self.assertEqual(differ.leaf_ratio(left, right), 0.9)
        # And one out of two children in common
        self.assertEqual(differ.child_ratio(left, right), 0.5)
        # But different id's, hence 0 as match
        self.assertEqual(differ.node_ratio(left, right), 0)

        # Here's the ones with the same id:
        left = differ.left.xpath('/document/story/section[1]')[0]
        right = differ.right.xpath('/document/story/section[2]')[0]

        # Only one out of two children in common
        self.assertEqual(differ.child_ratio(left, right), 0.5)
        # But same id's, hence 1 as match
        self.assertEqual(differ.node_ratio(left, right), 1.0)

        # The last ones are completely similar, but only one
        # has an xml:id, so they do not match.
        left = differ.left.xpath('/document/story/section[3]')[0]
        right = differ.right.xpath('/document/story/section[3]')[0]
        self.assertAlmostEqual(differ.leaf_ratio(left, right), 0.81818181818)
        self.assertEqual(differ.child_ratio(left, right), 1.0)
        self.assertEqual(differ.node_ratio(left, right), 0)

    def test_compare_node_rename(self):
        left = u"""<document>
  <para>First paragraph</para>
  <para attr="value">Second paragraph</para>
  <para attr="value">Third paragraph</para>
</document>
"""

        right = u"""<document>
  <section>First paragraph</section>
  <section attr="something else">Second paragraph</section>
  <section attr="something else">A different text</section>
</document>
"""

        differ = Differ()
        differ.set_trees(etree.fromstring(left), etree.fromstring(right))
        differ.match()

        # Make some choice comparisons here.
        left = differ.left.xpath('/document/para[1]')[0]
        right = differ.right.xpath('/document/section[1]')[0]

        # These have different tags, but should still match
        self.assertEqual(differ.leaf_ratio(left, right),
                         1.0)

        # These have different tags, and different attribute value,
        # but still similar enough
        left = differ.left.xpath('/document/para[2]')[0]
        right = differ.right.xpath('/document/section[2]')[0]

        # These have different tags, but should still match
        self.assertAlmostEqual(differ.leaf_ratio(left, right),
                               0.76190476190476)

        # These have different tags, and different attribute value,
        # but still similar enough
        left = differ.left.xpath('/document/para[3]')[0]
        right = differ.right.xpath('/document/section[3]')[0]

        # These are too different
        self.assertAlmostEqual(differ.leaf_ratio(left, right),
                               0.45161290322580)

    def test_compare_namespaces(self):
        left = u"""<document>
  <foo:para xmlns:foo="someuri">First paragraph</foo:para>
</document>
"""

        right = u"""<document>
  <foo:para xmlns:foo="otheruri">First paragraph</foo:para>
</document>
"""

        differ = Differ()
        differ.set_trees(etree.fromstring(left), etree.fromstring(right))
        differ.match()

        # Make some choice comparisons here.
        left = differ.left.xpath('/document/foo:para[1]',
                                 namespaces={'foo': 'someuri'})[0]
        right = differ.right.xpath('/document/foo:para[1]',
                                   namespaces={'foo': 'otheruri'})[0]

        # These have different namespaces, but should still match
        self.assertEqual(differ.leaf_ratio(left, right),
                         1.0)

    def test_different_ratio_modes(self):
        node1 = etree.Element('para')
        node1.text = "This doesn't match at all"
        node2 = etree.Element('para')
        node2.text = "It's completely different"
        node3 = etree.Element('para')
        node3.text = "Completely different from before"

        # These texts are very different
        differ = Differ(ratio_mode='accurate')
        self.assertAlmostEqual(differ.leaf_ratio(node1, node2), 0.24)
        # However, the quick_ratio doesn't catch that, and think they match
        differ = Differ(ratio_mode='fast')
        self.assertAlmostEqual(differ.leaf_ratio(node1, node2), 0.64)
        # It still realizes these sentences are different, though.
        differ = Differ(ratio_mode='fast')
        self.assertAlmostEqual(differ.leaf_ratio(node1, node3), 0.4561403508)
        # Faster thinks the first two are the same!
        differ = Differ(ratio_mode='faster')
        self.assertAlmostEqual(differ.leaf_ratio(node1, node2), 1.0)
        # And that the third is almost the same
        differ = Differ(ratio_mode='faster')
        self.assertAlmostEqual(differ.leaf_ratio(node1, node3), 0.8771929824)

        # Invalid modes raise error:
        with self.assertRaises(ValueError):
            differ = Differ(ratio_mode='allezlebleus')


class MatchTests(unittest.TestCase):

    def _match(self, left, right):
        left_tree = etree.fromstring(left)
        right_tree = etree.fromstring(right)
        differ = Differ()
        differ.set_trees(left_tree, right_tree)
        matches = differ.match()
        lpath = differ.left.getroottree().getpath
        rpath = differ.right.getroottree().getpath
        return [(lpath(item[0]), rpath(item[1])) for item in matches]

    def test_same_tree(self):
        xml = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section xml:id="oldlast" ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._match(xml, xml)
        nodes = list(utils.post_order_traverse(etree.fromstring(xml)))
        # Everything matches
        self.assertEqual(len(result), len(nodes))

    def test_no_xml_id_match(self):
        # Here we insert a section first, but because they contain numbering
        # it's easy to match section 1 in left with section 2 in right,
        # though it should be detected as an insert.

        # If the number of similar attributes are few it works fine, the
        # differing content of the ref="3" section means it's detected to
        # be an insert.
        left = u"""<document>
            <story firstPageTemplate="FirstPage">
            <section ref="3" single-ref="3">
            <para>First paragraph</para>
            </section>
            <section ref="4" single-ref="4">
            <para>Last paragraph</para>
            </section>
            </story>
    </document>
    """

        # We even detect that the first section is an insert without
        # xmlid, but that's less reliable.
        right = u"""<document>
            <story firstPageTemplate="FirstPage">
            <section ref="3" single-ref="3">
            <para>New paragraph</para>
            </section>
            <section ref="4" single-ref="4">
            <para>First paragraph</para>
            </section>
            <section ref="5" single-ref="5">
            <para>Last paragraph</para>
            </section>
            </story>
    </document>
    """

        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document/story/section[1]/para',
             '/document/story/section[2]/para'),
            ('/document/story/section[1]',
             '/document/story/section[2]'),
            ('/document/story/section[2]/para',
             '/document/story/section[3]/para'),
            ('/document/story/section[2]',
             '/document/story/section[3]'),
            ('/document/story',
             '/document/story'),
            ('/document',
             '/document')
        ])

    def test_with_xmlid(self):
        # This first section contains attributes that are similar (and longer
        # than the content text. That would trick the matcher into matching
        # the oldfirst and the newfirst section to match, except that we
        # this time also have xml:id's, and they trump everything else!
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3" xml:id="oldfirst"
                 description="This is to trick the differ">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4" xml:id="oldsecond">
            <para>Second paragraph</para>
        </section>
        <section ref="5" single-ref="5" xml:id="oldlast">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        # We even detect that the first section is an insert without
        # xmlid, but that's less reliable.
        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3" xml:id="newfirst"
                 description="This is to trick the differ">
            <para>New paragraph</para>
        </section>
        <section ref="4" single-ref="4" xml:id="oldfirst">
            <para>First paragraph</para>
        </section>
        <section ref="5" single-ref="5" xml:id="oldsecond">
            <para>Second paragraph</para>
        </section>
        <section ref="6" single-ref="6" xml:id="oldlast">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document/story/section[1]/para',
             '/document/story/section[2]/para'),
            ('/document/story/section[1]',
             '/document/story/section[2]'),
            ('/document/story/section[2]/para',
             '/document/story/section[3]/para'),
            ('/document/story/section[2]',
             '/document/story/section[3]'),
            ('/document/story/section[3]/para',
             '/document/story/section[4]/para'),
            ('/document/story/section[3]',
             '/document/story/section[4]'),
            ('/document/story',
             '/document/story'),
            ('/document',
             '/document')
        ])

    def test_change_attribs(self):

        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First</para>
        </section>
        <section xml:id="oldlast" ref="4" single-ref="4">
            <para>Last</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="4" single-ref="4">
            <para>First</para>
        </section>
        <section xml:id="oldlast" ref="5" single-ref="5">
            <para>Last</para>
        </section>
    </story>
</document>
"""
        # It matches everything straight, which means the attrib changes
        # should become updates, which makes sense.
        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document/story/section[1]/para',
             '/document/story/section[1]/para'),
            ('/document/story/section[1]',
             '/document/story/section[1]'),
            ('/document/story/section[2]/para',
             '/document/story/section[2]/para'),
            ('/document/story/section[2]',
             '/document/story/section[2]'),
            ('/document/story',
             '/document/story'),
            ('/document',
             '/document')
        ])

    def test_move_paragraph(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Second paragraph</para>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document/story/section[1]/para[1]',
             '/document/story/section[1]/para'),
            ('/document/story/section[1]/para[2]',
             '/document/story/section[2]/para[1]'),
            ('/document/story/section[1]', '/document/story/section[1]'),
            ('/document/story/section[2]/para',
             '/document/story/section[2]/para[2]'),
            ('/document/story/section[2]', '/document/story/section[2]'),
            ('/document/story', '/document/story'),
            ('/document', '/document')
        ])

    def test_match_complex_text(self):
        left = """<wrap id="1533728456.41"><para>
            Consultant shall not indemnify and hold Company, its
            affiliates and their respective directors,
            officers, agents and employees harmless from and
            against all claims, demands, losses, damages and
            judgments, including court costs and attorneys'
            fees, arising out of or based upon (a) any claim
            that the Services provided hereunder or, any
            related Intellectual Property Rights or the
            exercise of any rights in or to any Company-Related
            Development or Pre-Existing Development or related
            Intellectual Property Rights infringe on,
            constitute a misappropriation of the subject matter
            of, or otherwise violate any patent, copyright,
            trade secret, trademark or other proprietary right
            of any person or breaches any person's contractual
            rights; This is strange, but <b>true</b>.
            </para></wrap>"""

        right = """<wrap id="1533728456.41"><para>

            Consultant <i>shall not</i> indemnify and hold
            Company, its affiliates and their respective
            directors, officers, agents and employees harmless
            from and against all claims, demands, losses,
            excluding court costs and attorneys' fees, arising
            out of or based upon (a) any claim that the
            Services provided hereunder or, any related
            Intellectual Property Rights or the exercise of any
            rights in or to any Company-Related Development or
            Pre-Existing Development or related Intellectual
            Property Rights infringe on, constitute a
            misappropriation of the subject matter of, or
            otherwise violate any patent, copyright, trade
            secret, trademark or other proprietary right of any
            person or breaches any person's contractual rights;
            This is very strange, but <b>true</b>.

            </para></wrap>"""

        result = self._match(left, right)
        self.assertEqual(result, [
            ('/wrap/para/b', '/wrap/para/b'),
            ('/wrap/para', '/wrap/para'),
            ('/wrap', '/wrap')
        ])

    def test_match_insert_node(self):
        left = u'''<document title="insert-node">
  <story id="id">

  </story>
</document>
'''
        right = u'''<document title="insert-node">
  <story id="id">

    <h1>Inserted <i>Node</i></h1>

  </story>
</document>'''
        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document/story', '/document/story'),
            ('/document', '/document'),
        ])

    def test_entirely_different(self):
        left = u'''<document title="insert-node">
  <story id="id">

  </story>
</document>
'''
        right = u'''<document title="something else">
    <h1>Inserted <i>Node</i></h1>
</document>'''
        result = self._match(left, right)
        self.assertEqual(result, [
            ('/document', '/document'),
        ])


class FastMatchTests(unittest.TestCase):

    def _match(self, left, right, fast_match):
        left_tree = etree.fromstring(left)
        right_tree = etree.fromstring(right)
        differ = Differ(fast_match=fast_match)
        differ.set_trees(left_tree, right_tree)
        matches = differ.match()
        lpath = differ.left.getroottree().getpath
        rpath = differ.right.getroottree().getpath
        return [(lpath(item[0]), rpath(item[1])) for item in matches]

    def test_move_paragraph(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Second paragraph</para>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        # Same matches as the non-fast match test, but the matches are
        # a different order.
        slow_result = sorted(self._match(left, right, False))
        fast_result = sorted(self._match(left, right, True))
        self.assertEqual(slow_result, fast_result)

    def test_move_children(self):
        # Here the paragraphs are all so similar that that each paragraph
        # will match any other.
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>Second paragraph</para>
            <para>Last paragraph</para>
            <para>First paragraph</para>
        </section>
    </story>
</document>
"""
        # The slow match will match the nodes that match *best*, so it will
        # find that paragraphs have moved around.
        slow_result = sorted(self._match(left, right, False))
        self.assertEqual(slow_result, [
            ('/document', '/document'),
            ('/document/story', '/document/story'),
            ('/document/story/section', '/document/story/section'),
            ('/document/story/section/para[1]',
             '/document/story/section/para[3]'),
            ('/document/story/section/para[2]',
             '/document/story/section/para[1]'),
            ('/document/story/section/para[3]',
             '/document/story/section/para[2]')
        ])

        # But the fast match will just pick any that matches.
        fast_result = sorted(self._match(left, right, True))
        self.assertEqual(fast_result, [
            ('/document', '/document'),
            ('/document/story', '/document/story'),
            ('/document/story/section', '/document/story/section'),
            ('/document/story/section/para[1]',
             '/document/story/section/para[1]'),
            ('/document/story/section/para[2]',
             '/document/story/section/para[2]'),
            ('/document/story/section/para[3]',
             '/document/story/section/para[3]')
        ])


class UpdateNodeTests(unittest.TestCase):
    """Testing only the update phase of the diffing"""

    def _match(self, left, right):
        left_tree = etree.fromstring(left)
        right_tree = etree.fromstring(right)
        differ = Differ()
        differ.set_trees(left_tree, right_tree)
        matches = differ.match()
        steps = []
        for left, right, m in matches:
            steps.extend(differ.update_node_attr(left, right))
            steps.extend(differ.update_node_text(left, right))

        return steps

    def test_same_tree(self):
        xml = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section xml:id="oldlast" ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._match(xml, xml)
        # Everything matches
        self.assertEqual(result, [])

    def test_attribute_changes(self):
        left = u"""<root><node attr1="ohyeah" attr2="ohno" attr3="maybe" """\
               u"""attr0="del">The contained text</node>And a tail!</root>"""

        right = u"""<root><node attr4="ohyeah" attr2="uhhuh" attr3="maybe" """\
                u"""attr5="new">The new text</node>Also a tail!</root>"""

        result = self._match(left, right)

        self.assertEqual(
            result,
            [
                UpdateAttrib('/root/node[1]', 'attr2', 'uhhuh'),
                RenameAttrib('/root/node[1]', 'attr1', 'attr4'),
                InsertAttrib('/root/node[1]', 'attr5', 'new'),
                DeleteAttrib('/root/node[1]', 'attr0'),
                UpdateTextIn('/root/node[1]', 'The new text'),
                UpdateTextAfter('/root/node[1]', 'Also a tail!'),
            ]
        )


class AlignChildrenTests(unittest.TestCase):
    """Testing only the align phase of the diffing"""

    def _align(self, left, right):
        left_tree = etree.fromstring(left)
        right_tree = etree.fromstring(right)
        differ = Differ()
        differ.set_trees(left_tree, right_tree)
        matches = differ.match()
        steps = []
        for left, right, m in matches:
            steps.extend(differ.align_children(left, right))
        return steps

    def test_same_tree(self):
        xml = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section xml:id="oldfirst" ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section xml:id="oldlast" ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._align(xml, xml)
        # Everything matches
        self.assertEqual(result, [])

    def test_move_paragraph(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Second paragraph</para>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._align(left, right)
        # Everything matches
        self.assertEqual(result, [])

    def test_move_children(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>Second paragraph</para>
            <para>Last paragraph</para>
            <para>First paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._align(left, right)
        self.assertEqual(result,
                         [MoveNode('/document/story/section/para[1]',
                                   '/document/story/section[1]', 2)])


class DiffTests(unittest.TestCase):
    """Testing only the align phase of the diffing"""

    def _diff(self, left, right):
        parser = etree.XMLParser(remove_blank_text=True)
        left_tree = etree.fromstring(left, parser)
        right_tree = etree.fromstring(right, parser)
        differ = Differ()
        differ.set_trees(left_tree, right_tree)
        editscript = list(differ.diff())
        compare_elements(differ.left, differ.right)
        return editscript

    def test_process(self):
        left = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
            <para>Third paragraph</para>
        </section>
        <deleteme>
            <para>Delete it</para>
        </deleteme>
    </story>
</document>
"""

        right = u"""<document>
    <story firstPageTemplate="FirstPage">
        <section ref="3" single-ref="3">
            <para>First paragraph</para>
            <para>Second paragraph</para>
        </section>
        <section ref="4" single-ref="4">
            <para>Third paragraph</para>
            <para>Fourth paragraph</para>
        </section>
    </story>
</document>
"""
        result = self._diff(left, right)
        self.assertEqual(
            result,
            [
                InsertNode('/document/story[1]', 'section', 1),
                InsertAttrib('/document/story/section[2]', 'ref', '4'),
                InsertAttrib('/document/story/section[2]', 'single-ref', '4'),
                MoveNode('/document/story/section[1]/para[3]',
                         '/document/story/section[2]', 0),
                InsertNode('/document/story/section[2]', 'para', 1),
                UpdateTextIn('/document/story/section[2]/para[2]',
                             'Fourth paragraph'),
                DeleteNode('/document/story/deleteme/para[1]'),
                DeleteNode('/document/story/deleteme[1]'),
            ]
        )

    def test_needs_align(self):
        left = "<root><n><p>1</p><p>2</p><p>3</p></n><n><p>4</p></n></root>"
        right = "<root><n><p>2</p><p>4</p></n><n><p>1</p><p>3</p></n></root>"
        result = self._diff(left, right)
        self.assertEqual(
            result,
            [
                MoveNode('/root/n[1]', '/root[1]', 1),
                MoveNode('/root/n[2]/p[2]', '/root/n[1]', 0),
            ]
        )

    def test_no_root_match(self):
        left = '<root attr="val"><root><n><p>1</p><p>2</p><p>3</p></n>'\
            '<n><p>4</p></n></root></root>'
        right = '<root><n><p>2</p><p>4</p></n><n><p>1</p><p>3</p></n></root>'
        result = self._diff(left, right)
        self.assertEqual(
            result,
            [
                DeleteAttrib('/root[1]', 'attr'),
                MoveNode('/root/root/n[2]', '/root[1]', 0),
                MoveNode('/root/root/n[1]', '/root[1]', 1),
                MoveNode('/root/n[2]/p[2]', '/root/n[1]', 0),
                DeleteNode('/root/root[1]')
            ]
        )

    def test_rmldoc(self):
        here = os.path.split(__file__)[0]
        lfile = os.path.join(here, 'test_data', 'rmldoc.left.xml')
        rfile = os.path.join(here, 'test_data', 'rmldoc.right.xml')
        with open(lfile, 'rt', encoding='utf8') as infile:
            left = infile.read()
        with open(rfile, 'rt', encoding='utf8') as infile:
            right = infile.read()

        result = self._diff(left, right)
        self.assertEqual(
            result,
            [
                InsertNode(
                    '/document/story[1]',
                    '{http://namespaces.shoobx.com/application}section',
                    4),
                InsertAttrib(
                    '/document/story/app:section[4]', 'hidden', 'false'),
                InsertAttrib(
                    '/document/story/app:section[4]', 'name', 'sign'),
                InsertAttrib(
                    '/document/story/app:section[4]', 'ref', '3'),
                InsertAttrib(
                    '/document/story/app:section[4]', 'removed', 'false'),
                InsertAttrib(
                    '/document/story/app:section[4]', 'single-ref', '3'),
                InsertAttrib(
                 '/document/story/app:section[4]', 'title', 'Signing Bonus'),
                UpdateAttrib('/document/story/app:section[5]', 'ref', '4'),
                UpdateAttrib(
                    '/document/story/app:section[5]', 'single-ref', '4'),
                UpdateAttrib('/document/story/app:section[6]', 'ref', '5'),
                UpdateAttrib(
                    '/document/story/app:section[6]', 'single-ref', '5'),
                UpdateAttrib('/document/story/app:section[7]', 'ref', '6'),
                UpdateAttrib(
                    '/document/story/app:section[7]', 'single-ref', '6'),
                UpdateAttrib('/document/story/app:section[8]', 'ref', '7'),
                UpdateAttrib(
                    '/document/story/app:section[8]', 'single-ref', '7'),
                UpdateAttrib('/document/story/app:section[9]', 'ref', '8'),
                UpdateAttrib(
                    '/document/story/app:section[9]', 'single-ref', '8'),
                UpdateAttrib('/document/story/app:section[10]', 'ref', '9'),
                UpdateAttrib(
                    '/document/story/app:section[10]', 'single-ref', '9'),
                UpdateAttrib('/document/story/app:section[11]', 'ref', '10'),
                UpdateAttrib(
                    '/document/story/app:section[11]', 'single-ref', '10'),
                UpdateAttrib('/document/story/app:section[12]', 'ref', '11'),
                UpdateAttrib(
                    '/document/story/app:section[12]', 'single-ref', '11'),
                UpdateAttrib('/document/story/app:section[14]', 'ref', '12'),
                UpdateAttrib(
                    '/document/story/app:section[14]', 'single-ref', '12'),
                InsertNode(
                    '/document/story/app:section[4]',
                    '{http://namespaces.shoobx.com/application}term',
                    0),
                InsertAttrib(
                 '/document/story/app:section[4]/app:term[1]', 'name',
                 'sign_bonus'),
                InsertAttrib(
                    '/document/story/app:section[4]/app:term[1]', 'set', 'ol'),
                InsertNode('/document/story/app:section[4]', 'para', 1),
                UpdateTextIn(
                    '/document/story/app:section[1]/para[2]/'
                    'app:placeholder[1]',
                    'consectetur'),
                InsertNode(
                 '/document/story/app:section[4]/para[1]',
                 '{http://namespaces.shoobx.com/application}ref',
                 0),
                InsertAttrib(
                 '/document/story/app:section[4]/para/app:ref[1]', 'name',
                 'sign'),
                InsertAttrib(
                    '/document/story/app:section[4]/para/app:ref[1]',
                    '{http://namespaces.shoobx.com/preview}body',
                    '<Ref>'),
                UpdateTextIn(
                    '/document/story/app:section[4]/para/app:ref[1]', '3'),
                UpdateTextAfter(
                    '/document/story/app:section[4]/para/app:ref[1]', 'eu'),
                InsertNode('/document/story/app:section[4]/para[1]', 'u', 1),
                UpdateTextAfter(
                    '/document/story/app:section[4]/para/u[1]',
                    'ntum augue.\n\nAliquam nec tortor diam. Ph'),
                InsertNode(
                 '/document/story/app:section[4]/para[1]',
                 '{http://namespaces.shoobx.com/application}placeholder',
                 2),
                InsertAttrib(
                 '/document/story/app:section[4]/para/app:placeholder[1]',
                 'field',
                 'ol.sign_bonus_include_amt'),
                InsertAttrib(
                 '/document/story/app:section[4]/para/app:placeholder[1]',
                 'missing',
                 'Signing Bonus Amount'),
                UpdateTextAfter(
                    '/document/story/app:section[4]/para/app:placeholder[1]',
                    'asellus congue accumsan tempor. Donec vel risus se'
                ),
                UpdateTextIn(
                 '/document/story/app:section[5]/para/app:ref[1]',
                 '4'),
                UpdateTextIn(
                 '/document/story/app:section[6]/para/app:ref[1]',
                 '5'),
                UpdateTextIn(
                 '/document/story/app:section[7]/para/app:ref[1]',
                 '6'),
                UpdateTextIn(
                 '/document/story/app:section[8]/para/app:ref[1]',
                 '7'),
                UpdateTextIn(
                 '/document/story/app:section[9]/para/app:ref[1]',
                 '8'),
                UpdateTextIn(
                 '/document/story/app:section[10]/para/app:ref[1]',
                 '9'),
                UpdateTextIn(
                 '/document/story/app:section[11]/para/app:ref[1]',
                 '10'),
                UpdateTextIn(
                 '/document/story/app:section[12]/para/app:ref[1]',
                 '11'),
                InsertNode('/document/story/app:section[4]/para/u[1]', 'b', 0),
                UpdateTextIn(
                    '/document/story/app:section[4]/para/u/b[1]',
                    'ger nec ferme'),
            ]
        )

    def test_sbt_template(self):
        here = os.path.split(__file__)[0]
        lfile = os.path.join(here, 'test_data', 'sbt_template.left.xml')
        rfile = os.path.join(here, 'test_data', 'sbt_template.right.xml')
        with open(lfile, 'rt', encoding='utf8') as infile:
            left = infile.read()
        with open(rfile, 'rt', encoding='utf8') as infile:
            right = infile.read()

        result = self._diff(left, right)

        # Most lines get too long and flake8 complains because of this part:
        bm_bm_bm = '/metal:block/metal:block/metal:block'
        self.assertEqual(
            result,
            [
                InsertNode(
                    bm_bm_bm + '[1]',
                    '{http://namespaces.shoobx.com/application}section',
                    0),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]',
                    'allowCustom',
                    'False'),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]',
                    'hidden',
                    "advisor.payment_type == 'none'"),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]',
                    'name',
                    'payment'),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]',
                    'title',
                    'Payment'),
                InsertNode(
                    bm_bm_bm + '/app:section[1]',
                    '{http://xml.zope.org/namespaces/tal}if',
                    0),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]/tal:if[1]',
                    'condition',
                    "python: advisor.payment_type == 'stock_award'"),
                InsertNode(
                    bm_bm_bm + '/app:section[1]',
                    '{http://xml.zope.org/namespaces/tal}if',
                    1),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]/tal:if[2]',
                    'condition',
                    "python: advisor.payment_type == 'cash'"),
                InsertNode(
                    bm_bm_bm + '/app:section[1]',
                    '{http://xml.zope.org/namespaces/tal}if',
                    2),
                InsertAttrib(
                    bm_bm_bm + '/app:section[1]/tal:if[3]',
                    'condition',
                    "python: advisor.payment_type == 'stock_award_and_cash'"),
                InsertNode(
                    bm_bm_bm + '/app:section[1]/tal:if[1]',
                    'para',
                    0),
                UpdateTextIn(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para[1]',
                    '\n        A '),
                InsertNode(
                    bm_bm_bm + '/app:section[1]/tal:if[2]',
                    'para',
                    0),
                UpdateTextIn(
                    bm_bm_bm + '/app:section[1]/tal:if[2]/para[1]',
                    '\n        More text for diffing purposes\n      '),
                InsertNode(
                    bm_bm_bm + '/app:section[1]/tal:if[3]',
                    'para',
                    0),
                UpdateTextIn(
                    bm_bm_bm + '/app:section[1]/tal:if[3]/para[1]',
                    '\n        Lorem hipster ipso facto\n      '),
                InsertNode(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para[1]',
                    'i',
                    0),
                UpdateTextIn(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para/i[1]',
                    'whole'),
                UpdateTextAfter(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para/i[1]',
                    ' load of formatted text and '),
                InsertNode(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para[1]',
                    'br',
                    1),
                UpdateTextAfter(
                    bm_bm_bm + '/app:section[1]/tal:if[1]/para/br[1]',
                    ' other stuff.\n      '),
                DeleteNode(
                    bm_bm_bm + '/app:section[2]/tal:if/para/b[1]'),
                DeleteNode(
                    bm_bm_bm + '/app:section[2]/tal:if/para[1]'),
                DeleteNode(
                    bm_bm_bm + '/app:section[2]/tal:if[1]'),
                DeleteNode(
                    bm_bm_bm + '/app:section[2]')
            ]
        )

    def test_namespace(self):
        # Test changing nodes and attributes with namespaces
        left = u"""<document xmlns:app="someuri">
    <story app:foo="FirstPage">
        <app:section>
            <foo:para xmlns:foo="otheruri">Lorem ipsum dolor sit amet,
                consectetur adipiscing elit. Pellentesque feugiat metus quam.
                Suspendisse potenti. Vestibulum quis ornare felis,
                ac elementum sem.</foo:para>
            <app:para xmlns:foo="otheruri">Second paragraph</app:para>
            <app:para>Third paragraph</app:para>
            <app:para>
                Paragraph to tweak the matching of the section node
            </app:para>
            <app:para>
                By making many matching children
            </app:para>
            <app:para>
               Until the node matches properly.
            </app:para>
        </app:section>
    </story>
</document>
"""

        right = u"""<document xmlns:app="someuri">
    <story app:foo="FirstPage">
        <app:section>
            <app:para>Lorem ipsum dolor sit amet,
                consectetur adipiscing elit. Pellentesque feugiat metus quam.
                Suspendisse potenti. Vestibulum quis ornare felis,
                ac elementum sem.</app:para>
            <app:para>Second paragraph</app:para>
            <app:para app:attrib="value">Third paragraph</app:para>
            <app:para>
                Paragraph to tweak the matching of the section node
            </app:para>
            <app:para>
                By making many matching children
            </app:para>
            <app:para>
               Until the node matches properly.
            </app:para>
         </app:section>
    </story>
</document>
"""
        result = self._diff(left, right)
        self.assertEqual(
            result,
            [
                RenameNode(
                    '/document/story/app:section/foo:para[1]',
                    '{someuri}para'),
                InsertAttrib(
                    '/document/story/app:section/app:para[3]',
                    '{someuri}attrib', 'value'),
            ]
        )

    def test_multiple_tag_deletes(self):
        left = u"""<document title="delte-node-ul">
    <story id="id">

        <ul>
            <li>One</li>
            <li>Two</li>
            <li>Three</li>
        </ul>

    </story>
</document>"""

        right = u"""<document title="delte-node-ul">
    <story id="id">
    </story>
</document>"""

        result = self._diff(left, right)
        self.assertEqual(
            result,
            [UpdateTextIn('/document/story[1]', '\n    '),
             DeleteNode('/document/story/ul/li[3]'),
             DeleteNode('/document/story/ul/li[2]'),
             DeleteNode('/document/story/ul/li[1]'),
             DeleteNode('/document/story/ul[1]'),
             ]
        )

    def test_insert_comment(self):
        left = u"<doc><body>Something</body></doc>"
        right = u"<doc><!-- New comment! --><body>Something</body></doc>"

        result = self._diff(left, right)
        self.assertEqual(
            result,
            [InsertComment('/doc[1]', 0, ' New comment! ')]
        )

    def test_issue_21_default_namespaces(self):
        # When you have a default namespace you get "*" instead of the
        # expected "tag" in the XPath. This is how libxml does it,
        # and they say it has to be like that, so we document it.
        left = '<tag xmlns="ns">old</tag>'
        right = '<tag xmlns="ns">new</tag>'
        result = self._diff(left, right)
        self.assertEqual(result[0].node, '/*[1]')



import unittest

from lxml import etree
from xmldiff import utils


class TraverseTests(unittest.TestCase):

    def test_post_order(self):
        xml = u'''<document>
    <story firstPageTemplate='FirstPage'>
        <section xml:id='oldfirst' ref='3' single-ref='3'>
            <para>First paragraph</para>
        </section>
        <section xml:id='oldlast' ref='4' single-ref='4'>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
'''
        root = etree.fromstring(xml)
        tree = root.getroottree()
        res = [tree.getpath(x) for x in utils.post_order_traverse(root)]
        self.assertEqual(res, ['/document/story/section[1]/para',
                               '/document/story/section[1]',
                               '/document/story/section[2]/para',
                               '/document/story/section[2]',
                               '/document/story',
                               '/document'])

    def test_reverse_post_order(self):
        xml = u'''<document>
    <story firstPageTemplate='FirstPage'>
        <section xml:id='oldfirst' ref='3' single-ref='3'>
            <para>First paragraph</para>
        </section>
        <section xml:id='oldlast' ref='4' single-ref='4'>
            <para>Last paragraph</para>
        </section>
    </story>
</document>
'''
        root = etree.fromstring(xml)
        tree = root.getroottree()
        res = [tree.getpath(x) for x in
               utils.reverse_post_order_traverse(root)]
        self.assertEqual(res, ['/document/story/section[2]/para',
                               '/document/story/section[2]',
                               '/document/story/section[1]/para',
                               '/document/story/section[1]',
                               '/document/story',
                               '/document'])

    def test_breadth_first(self):
        xml = u'''<document>
    <story>
        <section>
            <para>First <i>paragraph</i></para>
            <para>Second paragraph</para>
        </section>
        <section>
            <para>Third paragraph</para>
            <para>Fourth <b>paragraph</b></para>
        </section>
    </story>
    <story>
        <section>
            <para>Fifth paragraph</para>
        </section>
    </story>
</document>
'''
        root = etree.fromstring(xml)
        tree = root.getroottree()
        res = [tree.getpath(x) for x in utils.breadth_first_traverse(root)]
        self.assertEqual(res, ['/document',
                               '/document/story[1]',
                               '/document/story[2]',
                               '/document/story[1]/section[1]',
                               '/document/story[1]/section[2]',
                               '/document/story[2]/section',
                               '/document/story[1]/section[1]/para[1]',
                               '/document/story[1]/section[1]/para[2]',
                               '/document/story[1]/section[2]/para[1]',
                               '/document/story[1]/section[2]/para[2]',
                               '/document/story[2]/section/para',
                               '/document/story[1]/section[1]/para[1]/i',
                               '/document/story[1]/section[2]/para[2]/b',
                               ])


class LongestCommonSubsequenceTests(unittest.TestCase):

    def _diff(self, left, right, result):
        res = []
        for x, y in utils.longest_common_subsequence(left, right):
            self.assertEqual(left[x], right[y])
            res.append(left[x])

        self.assertEqual(''.join(res), result)

    def test_lcs(self):

        self._diff('ABCDEF', 'ABCDEF', 'ABCDEF')

        self._diff('ABCDEF', 'GHIJKL', '')

        self._diff('ABCDEF', 'ACDQRB', 'ACD')

        self._diff('CXCDEFX', 'CDEFX', 'CDEFX')

        self._diff('HUMAN', 'CHIMPANZEE', 'HMAN')

        self._diff('ABCDEF', 'A', 'A')

        self._diff('123AAAAAAAAA', '123BBBBBBBBB', '123')

        self._diff('AAAAAAAAA123', 'BBBBBBBBB123', '123')

        self._diff('ABCDE1', '1FGHIJK', '1')

        # There are several correct options here, make sure that doesn't
        # confuse it, we want just one, and don't care which.
        self._diff('HORSEBACK', 'SNOWFLAKE', 'SAK')

        # Empty sequences:
        self._diff('', '', '')


class MakeAsciiTreeTests(unittest.TestCase):

    def test_make_ascii_tree(self):
        xml = u'''<document xmlns:diff="http://namespaces.shoobx.com/diff">
    <story firstPageTemplate='FirstPage'>
        <section xml:id='oldfirst' ref='3' single-ref='3'>
            <para diff:delete="">First paragraph</para>
        </section>
        <section xml:id='oldlast' ref='4' single-ref='4'>
            <para><diff:insert>Last paragraph</diff:insert></para>
        </section>
    </story>
</document>
'''
        root = etree.fromstring(xml)
        tree = utils.make_ascii_tree(root)
        self.assertEqual(
            tree,
            ' document \n   story \n     section \n       para (delete)\n'
            '     section \n       para \n         diff:insert '
        )



from io import open
from setuptools import setup, find_packages

version = '2.1'

with open('README.rst', 'rt', encoding='utf8') as readme:
    description = readme.read()

with open('CHANGES.rst', 'rt', encoding='utf8') as changes:
    history = changes.read()


setup(name='xmldiff',
      version=version,
      description="Creates diffs of XML files",
      long_description=description + '\n' + history,
      # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Topic :: Text Processing :: Markup :: XML',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'License :: OSI Approved :: MIT License',
                   ],
      keywords='xml html diff',
      author='Lennart Regebro',
      author_email='lregebro@shoobx.com',
      url='https://github.com/Shoobx/xmldiff',
      license='MIT',
      packages=find_packages(exclude=['doc', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'lxml>=3.1.0',
          'six',
      ],
      test_suite='tests',
      entry_points={
               'console_scripts': [
                   'xmldiff = xmldiff.main:run',
               ],
      },
)



# -*- coding: UTF-8 -*-
import os
import sys
import unittest

from lxml import etree
from xmldiff import diff, formatting, main

from .testing import generate_filebased_cases

START = u'<document xmlns:diff="http://namespaces.shoobx.com/diff"><node'
END = u'</node></document>'


class PlaceholderMakerTests(unittest.TestCase):

    def test_get_placeholder(self):
        replacer = formatting.PlaceholderMaker()
        # Get a placeholder:
        ph = replacer.get_placeholder(
            etree.Element('tag'), formatting.T_OPEN, None)
        self.assertEqual(ph, u'\ue005')
        # Do it again:
        ph = replacer.get_placeholder(
            etree.Element('tag'), formatting.T_OPEN, None)
        self.assertEqual(ph, u'\ue005')
        # Get another one
        ph = replacer.get_placeholder(
            etree.Element('tag'), formatting.T_CLOSE, ph)
        self.assertEqual(ph, u'\ue006')

    def test_do_element(self):
        replacer = formatting.PlaceholderMaker(['p'], ['b'])

        # Formatting tags get replaced, and the content remains
        text = u'<p>This is a tag with <b>formatted</b> text.</p>'
        element = etree.fromstring(text)
        replacer.do_element(element)

        self.assertEqual(
            etree.tounicode(element),
            u'<p>This is a tag with \ue006formatted\ue005 text.</p>')

        replacer.undo_element(element)
        self.assertEqual(etree.tounicode(element), text)

        # Non formatting tags get replaced with content
        text = u'<p>This is a tag with <foo>formatted</foo> text.</p>'
        element = etree.fromstring(text)
        replacer.do_element(element)
        result = etree.tounicode(element)
        self.assertEqual(
            result,
            u'<p>This is a tag with \ue007 text.</p>')

        # Single formatting tags still get two placeholders.
        text = u'<p>This is a <b/> with <foo/> text.</p>'
        element = etree.fromstring(text)
        replacer.do_element(element)
        result = etree.tounicode(element)
        self.assertEqual(
            result,
            u'<p>This is a \ue009\ue008 with \ue00a text.</p>')

    def test_do_undo_element(self):
        replacer = formatting.PlaceholderMaker(['p'], ['b'])

        # Formatting tags get replaced, and the content remains
        text = u'<p>This <is/> a <f>tag</f> with <b>formatted</b> text.</p>'
        element = etree.fromstring(text)
        replacer.do_element(element)

        self.assertEqual(
            element.text,
            u'This \ue005 a \ue006 with \ue008formatted'
            u'\ue007 text.')

        replacer.undo_element(element)
        result = etree.tounicode(element)
        self.assertEqual(result, text)

    def test_do_undo_element_double_format(self):
        replacer = formatting.PlaceholderMaker(['p'], ['b', 'u'])

        # Formatting tags get replaced, and the content remains
        text = u'<p>This is <u>doubly <b>formatted</b></u> text.</p>'
        element = etree.fromstring(text)
        replacer.do_element(element)

        self.assertEqual(
            element.text,
            u'This is \ue006doubly \ue008formatted\ue007'
            u'\ue005 text.')

        replacer.undo_element(element)
        result = etree.tounicode(element)
        self.assertEqual(result, text)

    def test_rml_bug(self):
        etree.register_namespace(formatting.DIFF_PREFIX, formatting.DIFF_NS)
        before_diff = u"""<document xmlns:diff="http://namespaces.shoobx.com/diff">
  <section>
    <para>
      <ref>4</ref>.
      <u><b>At Will Employment</b></u>
      .\u201cText\u201d
    </para>
  </section>
</document>"""
        tree = etree.fromstring(before_diff)
        replacer = formatting.PlaceholderMaker(
            text_tags=('para',), formatting_tags=('b', 'u', 'i',))
        replacer.do_tree(tree)
        after_diff = u"""<document xmlns:diff="http://namespaces.shoobx.com/diff">
  <section>
    <para>
      <insert>\ue005</insert>.
      \ue007\ue009At Will Employment\ue008\ue006
      .\u201c<insert>New </insert>Text\u201d
    </para>
  </section>
</document>"""

        # The diff formatting will find some text to insert.
        delete_attrib = u'{%s}delete-format' % formatting.DIFF_NS
        replacer.placeholder2tag[u'\ue006'
                                 ].element.attrib[delete_attrib] = ''
        replacer.placeholder2tag[u'\ue007'
                                 ].element.attrib[delete_attrib] = ''
        tree = etree.fromstring(after_diff)
        replacer.undo_tree(tree)
        result = etree.tounicode(tree)
        expected = u"""<document xmlns:diff="http://namespaces.shoobx.com/diff">
  <section>
    <para>
      <insert><ref>4</ref></insert>.
      <u diff:delete-format=""><b>At Will Employment</b></u>
      .\u201c<insert>New </insert>Text\u201d
    </para>
  </section>
</document>"""
        self.assertEqual(result, expected)

    def test_placeholder_overflow(self):
        # PY3: This whole test is Python 2 support.
        # Test what happens when we have more than 6400 placeholders,
        # by patching the placeholder:
        try:
            orig_start = formatting.PLACEHOLDER_START
            # This is the last character of the Private use area
            formatting.PLACEHOLDER_START = 0xF8FF

            replacer = formatting.PlaceholderMaker(['p'], ['b'])

            # Formatting tags get replaced, and the content remains
            text = u'<p>This <is/> a <f>tag</f> with <b>some</b> text.</p>'
            element = etree.fromstring(text)
            replacer.do_element(element)

            #
            self.assertEqual(
                element.text,
                u'This \uf904 a \uf905 with \uf907some'
                u'\uf906 text.')

            try:
                # If this is a wide build, also test what happens if we
                # get over 8192 substitutions, and overflow the 2-byte code.
                # (On narrow builds this will give an error)
                formatting.PLACEHOLDER_START = 0xFFFF

                replacer = formatting.PlaceholderMaker(['p'], ['b'])

                # Formatting tags get replaced, and the content remains
                text = u'<p>This <is/> a <f>tag</f> with <b>some</b> text.</p>'
                element = etree.fromstring(text)
                replacer.do_element(element)

                # This should raise an error on a narrow build
                self.assertEqual(
                    element.text,
                    u'This \U00010004 a \U00010005 with \U00010007some'
                    u'\U00010006 text.')
            except ValueError:
                if sys.maxunicode > 0x10000:
                    # This is a wide build, we should NOT get an error
                    raise

        finally:
            # Set it back
            formatting.PLACEHOLDER_START = orig_start


class XMLFormatTests(unittest.TestCase):

    def _format_test(self, left, action, expected):
        formatter = formatting.XMLFormatter(pretty_print=False)
        result = formatter.format([action], etree.fromstring(left))
        self.assertEqual(result, expected)

    def test_incorrect_xpaths(self):
        left = u'<document><node a="v"/><node>Text</node></document>'
        expected = START + u' diff:delete-attr="a">Text' + END

        with self.assertRaises(ValueError):
            action = diff.DeleteAttrib('/document/node', 'a')
            self._format_test(left, action, expected)

        with self.assertRaises(ValueError):
            action = diff.DeleteAttrib('/document/ummagumma', 'a')
            self._format_test(left, action, expected)

    def test_del_attr(self):
        left = u'<document><node a="v">Text</node></document>'
        action = diff.DeleteAttrib('/document/node', 'a')
        expected = START + u' diff:delete-attr="a">Text' + END

        self._format_test(left, action, expected)

    def test_del_node(self):
        left = u'<document><node attr="val">Text</node></document>'
        action = diff.DeleteNode('/document/node')
        expected = START + u' attr="val" diff:delete="">Text' + END

        self._format_test(left, action, expected)

    def test_del_text(self):
        left = u'<document><node attr="val">Text</node></document>'
        action = diff.UpdateTextIn('/document/node', None)
        expected = START + u' attr="val"><diff:delete>Text</diff:delete>' + END

        self._format_test(left, action, expected)

    def test_insert_attr(self):
        left = u'<document><node>We need more text</node></document>'
        action = diff.InsertAttrib('/document/node', 'attr', 'val')
        expected = START + u' attr="val" diff:add-attr="attr">'\
            u'We need more text' + END

        self._format_test(left, action, expected)

    def test_insert_node(self):
        left = u'<document></document>'
        action = diff.InsertNode('/document', 'node', 0)
        expected = START + u' diff:insert=""/></document>'

        self._format_test(left, action, expected)

    def test_move_attr(self):
        # The library currently only uses move attr for when attributes are
        # renamed:
        left = u'<document><node attr="val">Text</node></document>'
        action = diff.RenameAttrib('/document/node', 'attr', 'bottr')
        expected = START + u' bottr="val" diff:rename-attr="attr:bottr"'\
            u'>Text' + END

        self._format_test(left, action, expected)

    def test_move_node(self):
        # Move 1 down
        left = u'<document><node id="1" /><node id="2" /></document>'
        action = diff.MoveNode('/document/node[1]', '/document', 1)
        expected = START + u' id="1" diff:delete=""/><node id="2"/><node '\
            u'id="1" diff:insert=""/></document>'

        self._format_test(left, action, expected)

        # Move 2 up (same result, different diff)
        left = u'<document><node id="1" /><node id="2" /></document>'
        action = diff.MoveNode('/document/node[2]', '/document', 0)
        expected = START + u' id="2" diff:insert=""/><node id="1"/><node '\
            u'id="2" diff:delete=""/></document>'

        self._format_test(left, action, expected)

    def test_rename_node(self):
        left = u'<document><node><para>Content</para>Tail</node></document>'
        action = diff.RenameNode('/document/node[1]/para[1]', 'newtag')
        expected = START + u'><newtag diff:rename="para">Content'\
            '</newtag>Tail' + END

        self._format_test(left, action, expected)

    def test_update_attr(self):
        left = u'<document><node attr="val"/></document>'
        action = diff.UpdateAttrib('/document/node', 'attr', 'newval')
        expected = START + u' attr="newval" diff:update-attr="attr:val"/>'\
            u'</document>'

        self._format_test(left, action, expected)

    def test_update_text_in(self):
        left = u'<document><node attr="val"/></document>'
        action = diff.UpdateTextIn('/document/node', 'Text')
        expected = START + u' attr="val"><diff:insert>Text</diff:insert>' + END

        self._format_test(left, action, expected)

        left = u'<document><node>This is a bit of text, right' + END
        action = diff.UpdateTextIn('/document/node',
                                   'Also a bit of text, rick')
        expected = START + u'><diff:delete>This is</diff:delete><diff:insert>'\
            u'Also</diff:insert> a bit of text, ri<diff:delete>ght'\
            u'</diff:delete><diff:insert>ck</diff:insert>' + END

        self._format_test(left, action, expected)

    def test_update_text_after_1(self):
        left = u'<document><node/><node/></document>'
        action = diff.UpdateTextAfter('/document/node[1]', 'Text')
        expected = START + u'/><diff:insert>Text</diff:insert>'\
            u'<node/></document>'

        self._format_test(left, action, expected)

    def test_update_text_after_2(self):
        left = u'<document><node/>This is a bit of text, right</document>'
        action = diff.UpdateTextAfter('/document/node',
                                      'Also a bit of text, rick')
        expected = START + u'/><diff:delete>This is</diff:delete>'\
            u'<diff:insert>Also</diff:insert> a bit of text, ri<diff:delete>'\
            u'ght</diff:delete><diff:insert>ck</diff:insert></document>'

        self._format_test(left, action, expected)


class DiffFormatTests(unittest.TestCase):

    def _format_test(self, action, expected):
        formatter = formatting.DiffFormatter()
        result = formatter.format([action], None)
        self.assertEqual(result, expected)

    def test_del_attr(self):
        action = diff.DeleteAttrib('/document/node', 'a')
        expected = '[delete-attribute, /document/node, a]'
        self._format_test(action, expected)

    def test_del_node(self):
        action = diff.DeleteNode('/document/node')
        expected = '[delete, /document/node]'
        self._format_test(action, expected)

    def test_del_text(self):
        action = diff.UpdateTextIn('/document/node', None)
        expected = '[update-text, /document/node, null]'
        self._format_test(action, expected)

    def test_insert_attr(self):
        action = diff.InsertAttrib('/document/node', 'attr', 'val')
        expected = '[insert-attribute, /document/node, attr, "val"]'
        self._format_test(action, expected)

    def test_insert_node(self):
        action = diff.InsertNode('/document', 'node', 0)
        expected = '[insert, /document, node, 0]'
        self._format_test(action, expected)

    def test_rename_attr(self):
        action = diff.RenameAttrib('/document/node', 'attr', 'bottr')
        expected = '[move-attribute, /document/node, attr, bottr]'
        self._format_test(action, expected)

    def test_move_node(self):
        # Move 1 down
        action = diff.MoveNode('/document/node[1]', '/document', 1)
        expected = '[move, /document/node[1], /document, 1]'
        self._format_test(action, expected)

        # Move 2 up (same result, different diff)
        action = diff.MoveNode('/document/node[2]', '/document', 0)
        expected = '[move, /document/node[2], /document, 0]'

        self._format_test(action, expected)

    def test_rename_node(self):
        # Move 1 down
        action = diff.RenameNode('/document/node[1]', 'newtag')
        expected = '[rename, /document/node[1], newtag]'
        self._format_test(action, expected)

        # Move 2 up (same result, different diff)
        action = diff.MoveNode('/document/node[2]', '/document', 0)
        expected = '[move, /document/node[2], /document, 0]'

        self._format_test(action, expected)

    def test_update_attr(self):
        action = diff.UpdateAttrib('/document/node', 'attr', 'newval')
        expected = '[update-attribute, /document/node, attr, "newval"]'
        self._format_test(action, expected)

    def test_update_text_in(self):
        action = diff.UpdateTextIn('/document/node', 'Text')
        expected = '[update-text, /document/node, "Text"]'
        self._format_test(action, expected)

        action = diff.UpdateTextIn('/document/node',
                                   'Also a bit of text, "rick"')
        expected = '[update-text, /document/node, '\
            u'"Also a bit of text, \\"rick\\""]'
        self._format_test(action, expected)

    def test_update_text_after_1(self):
        action = diff.UpdateTextAfter('/document/node[1]', 'Text')
        expected = '[update-text-after, /document/node[1], "Text"]'
        self._format_test(action, expected)

    def test_update_text_after_2(self):
        action = diff.UpdateTextAfter('/document/node',
                                      'Also a bit of text, rick')
        expected = '[update-text-after, /document/node, '\
            u'"Also a bit of text, rick"]'
        self._format_test(action, expected)

    def test_insert_comment(self):
        action = diff.InsertComment('/document/node', 2, 'Commentary')
        expected = '[insert-comment, /document/node, 2, "Commentary"]'
        self._format_test(action, expected)


class XmlDiffFormatTests(unittest.TestCase):
    # RenameAttr and MoveNode requires an orig_tree, so they
    # are not tested in the _format_test tests, but in the
    # all_actions test, which uses test_data files.

    def _format_test(self, action, expected):
        formatter = formatting.XmlDiffFormatter()
        result = formatter.format([action], None)
        self.assertEqual(result, expected)

    def test_del_attr(self):
        action = diff.DeleteAttrib('/document/node', 'a')
        expected = '[remove, /document/node/@a]'
        self._format_test(action, expected)

    def test_del_node(self):
        action = diff.DeleteNode('/document/node')
        expected = '[remove, /document/node]'
        self._format_test(action, expected)

    def test_del_text(self):
        action = diff.UpdateTextIn('/document/node', None)
        expected = '[update, /document/node/text()[1], null]'
        self._format_test(action, expected)

    def test_insert_attr(self):
        action = diff.InsertAttrib('/document/node', 'attr', 'val')
        expected = '[insert, /document/node, \n<@attr>\nval\n</@attr>]'
        self._format_test(action, expected)

    def test_insert_node(self):
        action = diff.InsertNode('/document', 'node', 0)
        expected = '[insert-first, /document, \n<node/>]'
        self._format_test(action, expected)

    def test_rename_node(self):
        # Move 1 down
        action = diff.RenameNode('/document/node[1]', 'newtag')
        expected = '[rename, /document/node[1], newtag]'
        self._format_test(action, expected)

        # Move 2 up (same result, different diff)
        action = diff.MoveNode('/document/node[2]', '/document', 0)
        expected = '[move-first, /document/node[2], /document]'
        self._format_test(action, expected)

    def test_update_attr(self):
        action = diff.UpdateAttrib('/document/node', 'attr', 'newval')
        expected = '[update, /document/node/@attr, "newval"]'
        self._format_test(action, expected)

    def test_update_text_in(self):
        action = diff.UpdateTextIn('/document/node', 'Text')
        expected = '[update, /document/node/text()[1], "Text"]'
        self._format_test(action, expected)

        action = diff.UpdateTextIn('/document/node',
                                   'Also a bit of text, "rick"')
        expected = '[update, /document/node/text()[1], '\
            u'"Also a bit of text, \\"rick\\""]'
        self._format_test(action, expected)

    def test_update_text_after_1(self):
        action = diff.UpdateTextAfter('/document/node[1]', 'Text')
        expected = '[update, /document/node[1]/text()[2], "Text"]'
        self._format_test(action, expected)

    def test_update_text_after_2(self):
        action = diff.UpdateTextAfter('/document/node',
                                      'Also a bit of text, rick')
        expected = '[update, /document/node/text()[2], '\
            u'"Also a bit of text, rick"]'
        self._format_test(action, expected)

    def test_all_actions(self):
        here = os.path.split(__file__)[0]
        lfile = os.path.join(here, 'test_data', 'all_actions.left.xml')
        rfile = os.path.join(here, 'test_data', 'all_actions.right.xml')

        formatter = formatting.XmlDiffFormatter()
        result = main.diff_files(lfile, rfile, formatter=formatter)
        expected = (
            u'[move-after, /document/node[2], /document/tag[1]]\n'
            u'[insert-comment, /document[1], 0,  Insert a new comment ]\n'
            u'[update, /document/node[1]/@name, "was updated"]\n'
            u'[remove, /document/node[1]/@attribute]\n'
            u'[insert, /document/node[1], \n'
            u'<@newtribute>\n'
            u'renamed\n'
            u'</@newtribute>]\n'
            u'[insert, /document/node[1], \n'
            u'<@this>\n'
            u'is new\n'
            u'</@this>]\n'
            u'[remove, /document/node[1]/@attr]\n'
            u'[update, /document/node[1]/text()[1], "\\n    Modified\\n  "]\n'
            u'[update, /document/node[1]/text()[2], "\\n    '
            u'New tail content\\n  "]\n'
            u'[rename, /document/node[2], nod]\n'
            u'[insert-after, /document/tail[1], \n'
            u'<new/>]\n'
            u'[remove, /document/tail[1]]'
        )
        self.assertEqual(result, expected)


class FormatterFileTests(unittest.TestCase):

    formatter = None  # Override this
    maxDiff = None

    def process(self, left, right):
        return main.diff_files(left, right, formatter=self.formatter)


class XMLFormatterFileTests(FormatterFileTests):

    # The XMLFormatter has no text or formatting tags, so
    formatter = formatting.XMLFormatter(pretty_print=False,
                                        normalize=formatting.WS_TEXT)


# Also test the bits that handle text tags:

class HTMLFormatterFileTests(FormatterFileTests):

    # We use a few tags for the placeholder tests.
    # <br/> is intentionally left out, to test an edge case
    # with empty non-formatting tags in text.
    formatter = formatting.XMLFormatter(
        normalize=formatting.WS_BOTH,
        pretty_print=True,
        text_tags=('p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'li'),
        formatting_tags=('b', 'u', 'i', 'strike', 'em', 'super',
                         'sup', 'sub', 'link', 'a', 'span'))


# Add tests that use no placeholder replacement (ie plain XML)
data_dir = os.path.join(os.path.dirname(__file__), 'test_data')
generate_filebased_cases(data_dir, XMLFormatterFileTests)

# Add tests that use placeholder replacement (ie HTML)
data_dir = os.path.join(os.path.dirname(__file__), 'test_data')
generate_filebased_cases(data_dir, HTMLFormatterFileTests, suffix='html')



"""All major API points and command-line tools"""
import pkg_resources

from argparse import ArgumentParser, FileType
from lxml import etree
from xmldiff import diff, formatting

__version__ = pkg_resources.require("xmldiff")[0].version

FORMATTERS = {
    'diff': formatting.DiffFormatter,
    'xml': formatting.XMLFormatter,
    'old': formatting.XmlDiffFormatter,
}


def diff_trees(left, right, diff_options=None, formatter=None):
    """Takes two lxml root elements or element trees"""
    if formatter is not None:
        formatter.prepare(left, right)
    if diff_options is None:
        diff_options = {}
    differ = diff.Differ(**diff_options)
    diffs = differ.diff(left, right)

    if formatter is None:
        return list(diffs)

    return formatter.format(diffs, left)


def _diff(parse_method, left, right, diff_options=None, formatter=None):
    normalize = bool(getattr(formatter, 'normalize', 1) & formatting.WS_TAGS)
    parser = etree.XMLParser(remove_blank_text=normalize)
    left_tree = parse_method(left, parser)
    right_tree = parse_method(right, parser)
    return diff_trees(left_tree, right_tree, diff_options=diff_options,
                      formatter=formatter)


def diff_texts(left, right, diff_options=None, formatter=None):
    """Takes two Unicode strings containing XML"""
    return _diff(etree.fromstring, left, right,
                 diff_options=diff_options, formatter=formatter)


def diff_files(left, right, diff_options=None, formatter=None):
    """Takes two filenames or streams, and diffs the XML in those files"""
    return _diff(etree.parse, left, right,
                 diff_options=diff_options, formatter=formatter)


def make_parser():
    parser = ArgumentParser(description='Create a diff for two XML files.',
                            add_help=False)
    parser.add_argument('file1', type=FileType('r'),
                        help='The first input file.')
    parser.add_argument('file2', type=FileType('r'),
                        help='The second input file.')
    parser.add_argument('-h', '--help', action='help',
                        help='Show this help message and exit.')
    parser.add_argument('-v', '--version', action='version',
                        help='Display version and exit.',
                        version='xmldiff %s' % __version__)
    parser.add_argument('-f', '--formatter', default='diff',
                        choices=list(FORMATTERS.keys()),
                        help='Formatter selection.')
    parser.add_argument('-w', '--keep-whitespace', action='store_true',
                        help='Do not strip ignorable whitespace.')
    parser.add_argument('-p', '--pretty-print', action='store_true',
                        help='Try to make XML output more readable.')
    parser.add_argument('-F', type=float,
                        help='A value betwen 0 and 1 that determines how '
                        'similar nodes must be to match.')
    parser.add_argument('--unique-attributes', type=str, nargs='?',
                        default='{http://www.w3.org/XML/1998/namespace}id',
                        help='A comma separated list of attributes '
                             'that uniquely identify a node. Can be empty.')
    parser.add_argument('--ratio-mode', default='fast',
                        choices={'accurate', 'fast', 'faster'},
                        help='Choose the node comparison optimization.')
    parser.add_argument('--fast-match', action='store_true',
                        help='A faster, less optimal match run.')
    return parser


def run(args=None):
    parser = make_parser()
    args = parser.parse_args(args=args)

    if args.keep_whitespace:
        normalize = formatting.WS_NONE
    else:
        normalize = formatting.WS_BOTH

    formatter = FORMATTERS[args.formatter](normalize=normalize,
                                           pretty_print=args.pretty_print)

    if args.unique_attributes is None:
        uniqueattrs = []
    else:
        uniqueattrs = args.unique_attributes.split(',')

    diff_options = {'ratio_mode': args.ratio_mode,
                    'F': args.F,
                    'fast_match': args.fast_match,
                    'uniqueattrs': uniqueattrs,
                    }
    result = diff_files(args.file1, args.file2, diff_options=diff_options,
                        formatter=formatter)
    print(result)
